/*  
 * GSM.cpp - Libary for using the GSM and GPS-Modul on the GSM/GPRS/GPS-Shield Rev.7
 * Initialization code and constants for talking to the SCI16IS750 are taken from the original
 * library from antrax datentechnik gmbh (http://antrax.de)
 * 
 * Libary is set to work with the arduino mega and debugs on Serial0.
 * This can be changed by changing defines in GPS.h
 */

#include <Arduino.h>
#include "GSM.h"

#define GSM_COMMAND(command) GSM_SERIAL.print(command);
                                                        // Arduino Duemilanove (ATmega328) & Arduino UNO (ATmega328)
GSM::GSM()                                
{         
 _debug = 0;     
 pinMode(GSM_POWER_PIN, OUTPUT);                                                 
}
GSM::GSM(byte dbg)                                
{              
 _debug = dbg;
 pinMode(GSM_POWER_PIN, OUTPUT);                                                 
}

byte always_match(char * trigger_line, char * input_line){
	return 1;
}
byte match_trailing_result(char * trigger_line, char * input_line){
  // check if input line ends with trigger line
  int trigger_len = strlen(trigger_line);
  int input_len = strlen(input_line);
  
  if(trigger_len > input_len){
    return 0; // no match, to short
  }  
  // start 1 = 1 to avoid check for empty 0
  for (int i = 0;i< trigger_len;i++){
    if(trigger_line[trigger_len - i] != input_line[input_len - i]){
      // not matching. 
      return 0;
    }
  }
  // all chars match
  return 1; 
}

bool GSM::initialize(char *pwd) { 
  State_Machine init_sm;
  byte size_at_command=sizeof("AT+CPIN=");
	
  char pin_setting_command[4 + size_at_command + 2]; // 4 pin chars, command chars, \r and 0
  strcpy (pin_setting_command,"AT+CPIN=");
  strcpy (pin_setting_command + size_at_command - 1,pwd); // write over intermedita 0
  strcpy (pin_setting_command + size_at_command + 3,"\r"); // write over inttermediate 0
  
  GSM_SERIAL.end();
 
  digitalWrite(GSM_POWER_PIN, LOW);                                              // disable Power GSM-Modul
  delay(1000);
  digitalWrite(GSM_POWER_PIN, HIGH);                                             // enable Power GSM-Modul

  GSM_SERIAL.begin(GSM_SERIAL_BAUD);
  
  buffer_pos=0;
  
  delay(3000);
  
  // run init state_machine
  // array of commands. index in array is = current state. append \r allways. ommit 1st state.
  char * init_commands[] = {"","AT\r","ATE0\r","AT+IPR=9600\r","AT#SIMDET=1\r","AT+CPIN?\r",pin_setting_command,"AT+CREG?\r"};
  init_sm.add_transition(1,2,"OK",match_trailing_result);
  init_sm.add_transition(2,3,"OK",match_trailing_result);
  init_sm.add_transition(3,4,"OK",match_trailing_result);
  init_sm.add_transition(4,5,"OK",match_trailing_result);
  init_sm.add_transition(5,6,"SIM PIN",match_trailing_result); // PIN not set
  init_sm.add_transition(5,5,"ERROR",match_trailing_result); // retry on PIN ERROR
  init_sm.add_transition(5,7,"READY",match_trailing_result); // PIN not set
  init_sm.add_transition(6,7,"OK",match_trailing_result); 
  init_sm.set_state(1);
  byte last_state = 1;
  
  byte command_sent = 0;
  byte retries =0;
  byte abort = 0;
    
  while((init_sm.get_state() < 7) && (!abort)) {
	
	if(command_sent == 0){
		char * command = init_commands[init_sm.get_state()];
		DEBUG_MSG("STATE:")
	    DEBUG_MSG(init_sm.get_state())
		DEBUG_MSG("Command:")
		DEBUG_MSG(command)
 
		send_command(command);
		command_sent = 1;
	}
  
	if(read_gsm_buffer()){
		// wir haben eine Antwort bekommen
		// prüfen ob neuer Zustand
		DEBUG_MSG("AW:")
		DEBUG_MSG(get_gsm_buffer())
		last_state = init_sm.get_state();
		if(init_sm.match_line(get_gsm_buffer())){// Zustandsübergang hat stattgefunden, match gefunden
			// ist der alte zustand = dem neuen?
			if(last_state == init_sm.get_state()){
				// ja. zuviele retries?
				if(++retries > MAX_RETRY_AT_COMMAND){
					abort = 1;
				} else {
					delay (500); // 500ms warten bis neuer versuch					
				}
			} else {
				// neuer zustand reset retry auf 0
				retries = 0;
			}
			// im nächsten loop passendes commando senden!
			command_sent = 0;
		} else {
			if(++retries > MAX_RETRY_AT_COMMAND){
					abort = 1;
			}
		}
	} // ende des reads und zustandsübergang
  }
  delay(1000); // give SIM time to setup
  return !abort;
}


               
bool GSM::send_SMS(char * phone_number, char * message){
  State_Machine sms_sm;
  
  // puffer sauber machen von alten Antworten
  clear_UART();

  // zusammenbauen des Strings für die SMS
  byte size_at_command = strlen("AT+CMGS=");
  byte size_message = strlen(message);
  byte size_phone_nr = strlen(phone_number);

  char  terminator[] = {26,0};
  char sms_command[250]; // 
  char message_terminated[250];
  strcpy (sms_command,"AT+CMGS=");
  strcpy (sms_command + size_at_command,phone_number); 
  strcpy (sms_command + size_at_command + size_phone_nr,"\r"); 
  strcpy (message_terminated,message); 
  strcpy (message_terminated + size_message, terminator); 

  DEBUG_MSG("SMS KOMMANDO:")
  DEBUG_MSG(sms_command)
  
  // setup zustandsmaschine
  char end_string[] = {26,'\r',0};
  char * init_commands[] = {"","AT\r","AT+CMGF=1\r",sms_command,message_terminated,"AT\r"};
  sms_sm.add_transition(1,2,"",always_match);
  sms_sm.add_transition(2,3,"OK",match_trailing_result);
  sms_sm.add_transition(3,4,">",match_trailing_result);
  sms_sm.add_transition(3,3,"ERROR: 314",match_trailing_result); 
  sms_sm.add_transition(3,3,"ERROR: 28",match_trailing_result); 
  sms_sm.add_transition(4,5,">",match_trailing_result);
  sms_sm.add_transition(3,3,"ERROR",match_trailing_result);
  sms_sm.add_transition(4,5,"OK",match_trailing_result);
  /* TO be done
  sms_sm.add_transition(5,6,"",always_match);
  */

  sms_sm.set_state(1);
  byte command_sent = 0;
  byte last_state = 1;
  byte retries =0;
  byte abort = 0;
  
  while((sms_sm.get_state() < 5) && (!abort)) {
	delay(100); // wait 100 ms between commands
	if(command_sent == 0){
		char * command = init_commands[sms_sm.get_state()];
		DEBUG_MSG("STATE:")
		DEBUG_MSG(sms_sm.get_state())
		DEBUG_MSG("Command:")
		DEBUG_MSG(command)
 
		send_command(command);
		command_sent = 1;
	}
  
	if(read_gsm_buffer()){
		// wir haben eine Antwort bekommen
		// prüfen ob neuer Zustand
		DEBUG_MSG("AW:")
		DEBUG_MSG(get_gsm_buffer())
		last_state = sms_sm.get_state();
		if(sms_sm.match_line(get_gsm_buffer())){// Zustandsübergang hat stattgefunden, match gefunden
			// ist der alte zustand = dem neuen?
			if(last_state == sms_sm.get_state()){
				// ja. zuviele retries?
				if(++retries > MAX_RETRY_AT_COMMAND){
					abort = 1;
				} else {
					delay (1000); // 1s warten bis neuer versuch					
				}
			} else {
				// neuer zustand reset retry auf 0
				retries = 0;
			}
			// im nächsten loop passendes commando senden!
			command_sent = 0;
		} else { // no match in SM
			abort = 1;
	    }
	}
  }
  return (!abort); // end with other than 4 is an error
}
// returns true when buffer is valid (once)
// removes trailing <CR> and <LF>
bool GSM::read_gsm_buffer(){
	if (GSM_SERIAL.available()) {
		byte tmp = GSM_SERIAL.read();
		gsm_buffer[buffer_pos++] = tmp;
		
		// check if we have reached the end of the answer and we did receive an answer
		if(buffer_pos > 2) { // more received than leading <LF><CR>
			if(gsm_buffer[buffer_pos-2] == '\r' && gsm_buffer[buffer_pos-1] == '\n'){
				gsm_buffer[buffer_pos-2] = 0; // terminate string before trailing 
				buffer_pos = 0;
				// ignore comments
				if(gsm_buffer[2] == '#'){
					return false;
				}
				return true;
			}
		}
	}		
	return false;
}

char * GSM::get_gsm_buffer(){
	return gsm_buffer;
}

void GSM::send_command(char * command){
	GSM_SERIAL.print(command);
}


char GSM::get_gsm_state(void) { 
	send_command("AT+CREG?\r");
	// blocking right now until response arives
	while(!read_gsm_buffer());
	char * response = get_gsm_buffer();
	
	// the status is after the first ','
	int pos = 0;
	while(response[pos++] != ',');
	return response[pos];
}

void GSM::clear_UART(){
	while(read_gsm_buffer()) ;
}